from pathlib import Path
from typing import TypedDict, Union

from ..properties.isvc.mlserver import (
    LightGBMProperties,
    MLServerProperties,
    SKLearnProperties,
    XGBoostProperties,
)
from ..properties.isvc.tritonserver import (
    ONNXProperties,
    TensorflowSavedModelProperties,
    TorchscriptProperties,
)
from ..resources.isvc.types import ModelType, ServingFrameworkType
from ..resources.types import ResourceType
from ..template import Template
from .utils import has_handler_class


class InferenceServiceValidator:
    def __init__(self, template: Template) -> None:
        self.template = template

    @staticmethod
    def validate_saved_model(props: TensorflowSavedModelProperties) -> None:
        class RequiredFile(TypedDict):
            name: str
            is_file: bool

        required_files: list[RequiredFile] = [
            {"name": "saved_model.pb", "is_file": True},
            {"name": "variables", "is_file": False},
            {"name": "fingerprint.pb", "is_file": True},
        ]
        for req in required_files:
            path_res: Path = props.SavedModelUri / req["name"]
            if req["is_file"]:
                if path_res.is_file() is False:
                    raise ValueError(f'File "{path_res}" not found or is not a file.')
            else:
                if path_res.is_dir() is False:
                    raise ValueError(f'Directory "{path_res}" not found or is not a directory.')

    @staticmethod
    def validate_model_uri(
        props: Union[TorchscriptProperties, ONNXProperties, XGBoostProperties, LightGBMProperties, SKLearnProperties],
    ) -> None:
        if props.ModelUri.is_file() is False:
            raise ValueError(f"File {props.ModelUri} not found or is not a file.")

    @staticmethod
    def validate_handler(props: MLServerProperties) -> None:
        # Check the handler script exists
        handler_script = "/".join(props.Handler.split(".")[:-1]) + ".py"
        script_path = props.CodeUri / handler_script
        if script_path.is_file() is False:
            raise ValueError(f"File {script_path} not found or is not a file.")

        # Check the handler entrypoint exists in the handler script
        cls_name = props.Handler.split(".")[-1]
        if has_handler_class(script_path, cls_name) is False:
            raise ValueError(f'Handler "{cls_name}" not found in {script_path}')

    def __call__(self):
        for _, resource in self.template.Resources.items():
            if resource.SuperType != ResourceType.INFERENCE_SERVICE:
                raise NotImplementedError(f"Resource {resource.Type.value} is not supported")

            if resource.ServingFrameworkType == ServingFrameworkType.TRITONSERVER and resource.ModelType in (
                ModelType.TORCHSCRIPT,
                ModelType.ONNX,
            ):
                self.validate_model_uri(resource.Properties)
            elif (
                resource.ServingFrameworkType == ServingFrameworkType.TRITONSERVER
                and resource.ModelType == ModelType.TENSORFLOW_SAVEDMODEL
            ):
                self.validate_saved_model(resource.Properties)
            elif resource.ServingFrameworkType == ServingFrameworkType.ML_SERVER and resource.ModelType in (
                ModelType.XGBOOST,
                ModelType.LIGHTGBM,
                ModelType.SKLEARN,
            ):
                self.validate_model_uri(resource.Properties)
                self.validate_handler(resource.Properties)
            else:
                raise NotImplementedError(f"Resource {resource.Type.value} is not supported")



from enum import Enum

from pydantic import BaseModel


class DataTypeEnum(str, Enum):
    BOOL = "TYPE_BOOL"
    UINT8 = "TYPE_UINT8"
    UINT16 = "TYPE_UINT16"
    UINT32 = "TYPE_UINT32"
    UINT64 = "TYPE_UINT64"
    INT8 = "TYPE_INT8"
    INT16 = "TYPE_INT16"
    INT32 = "TYPE_INT32"
    INT64 = "TYPE_INT64"
    FP16 = "TYPE_FP16"
    FP32 = "TYPE_FP32"
    FP64 = "TYPE_FP64"
    STRING = "TYPE_STRING"
    BF16 = "TYPE_BF16"


class BaseSignature(BaseModel):
    DataType: DataTypeEnum
    Dims: list[int]
    Name: str


class InputSignature(BaseSignature):
    MonitoringElement: str


class OutputSignature(BaseSignature):
    pass



from pathlib import Path
from typing import Optional

from ..base import InferenceServiceProperties
from .types import PlatformType


class TritonserverProperties(InferenceServiceProperties):
    MaxBatchSize: int
    ModelRepositoryUri: Optional[Path] = None
    Platform: PlatformType

    def __setattr__(self, name, value):
        if name == "Platform":
            raise AttributeError("Platform is a constant and cannot be changed.")
        super().__setattr__(name, value)



from pathlib import Path

import pytest

from dism_core.properties.isvc.mlserver import LightGBMProperties, SKLearnProperties, XGBoostProperties
from dism_core.properties.isvc.signature import DataTypeEnum
from dism_core.properties.isvc.tritonserver import (
    ONNXProperties,
    TensorflowSavedModelProperties,
    TorchscriptProperties,
)
from dism_core.properties.isvc.tritonserver.types import PlatformType


@pytest.fixture
def valid_torchcript_properties() -> dict:
    return {
        "Description": "Example description",
        "Name": "autoencoder",
        "MetricKey": "output_1",
        "BuiltinThreshold": False,
        "AnomalyThreshold": 0.1,
        "InputSignature": [
            {
                "Name": "input_0",
                "MonitoringElement": "JetMET/MET/pfMETT1/Cleaned/METSig",
                "DataType": "TYPE_FP32",
                "Dims": [-1, 51],
            }
        ],
        "OutputSignature": [
            {"Name": "output_0", "DataType": "TYPE_FP32", "Dims": [-1, 51]},
            {"Name": "output_1", "DataType": "TYPE_FP32", "Dims": [-1]},
        ],
        "MaxBatchSize": 0,
        "ModelUri": "example/model.pt",
    }


@pytest.fixture
def valid_tensorflow_savedmodel_properties() -> dict:
    return {
        "Description": "Example description",
        "Name": "autoencoder",
        "MetricKey": "output_1",
        "BuiltinThreshold": False,
        "AnomalyThreshold": 0.1,
        "InputSignature": [
            {
                "Name": "input_0",
                "MonitoringElement": "JetMET/MET/pfMETT1/Cleaned/METSig",
                "DataType": "TYPE_FP32",
                "Dims": [-1, 51],
            }
        ],
        "OutputSignature": [
            {"Name": "output_0", "DataType": "TYPE_FP32", "Dims": [-1, 51]},
            {"Name": "output_1", "DataType": "TYPE_FP32", "Dims": [-1]},
        ],
        "MaxBatchSize": 0,
        "SavedModelUri": "example/saved_model",
    }


@pytest.fixture
def valid_onnx_properties() -> dict:
    return {
        "Description": "Example description",
        "Name": "autoencoder",
        "MetricKey": "output_1",
        "BuiltinThreshold": False,
        "AnomalyThreshold": 0.1,
        "InputSignature": [
            {
                "Name": "input_0",
                "MonitoringElement": "JetMET/MET/pfMETT1/Cleaned/METSig",
                "DataType": "TYPE_FP32",
                "Dims": [-1, 51],
            }
        ],
        "OutputSignature": [
            {"Name": "output_0", "DataType": "TYPE_FP32", "Dims": [-1, 51]},
            {"Name": "output_1", "DataType": "TYPE_FP32", "Dims": [-1]},
        ],
        "MaxBatchSize": 0,
        "ModelUri": "example/model.onnx",
    }


@pytest.fixture
def valid_xgboost_properties() -> dict:
    return {
        "Description": "Example description",
        "Name": "autoencoder",
        "MetricKey": "output_1",
        "BuiltinThreshold": False,
        "AnomalyThreshold": 0.1,
        "InputSignature": [
            {
                "Name": "input_0",
                "MonitoringElement": "JetMET/MET/pfMETT1/Cleaned/METSig",
                "DataType": "TYPE_FP32",
                "Dims": [-1, 51],
            }
        ],
        "OutputSignature": [
            {"Name": "output_0", "DataType": "TYPE_FP32", "Dims": [-1, 51]},
            {"Name": "output_1", "DataType": "TYPE_FP32", "Dims": [-1]},
        ],
        "CodeUri": "example/xgboost",
        "ModelUri": "example/xgboost/0001.ubj",
        "Handler": "app.Handler",
    }


@pytest.fixture
def valid_lightgbm_properties() -> dict:
    return {
        "Description": "Example description",
        "Name": "autoencoder",
        "MetricKey": "output_1",
        "BuiltinThreshold": False,
        "AnomalyThreshold": 0.1,
        "InputSignature": [
            {
                "Name": "input_0",
                "MonitoringElement": "JetMET/MET/pfMETT1/Cleaned/METSig",
                "DataType": "TYPE_FP32",
                "Dims": [-1, 51],
            }
        ],
        "OutputSignature": [
            {"Name": "output_0", "DataType": "TYPE_FP32", "Dims": [-1, 51]},
            {"Name": "output_1", "DataType": "TYPE_FP32", "Dims": [-1]},
        ],
        "CodeUri": "example/lightgbm",
        "ModelUri": "example/lightgbm/0001.txt",
        "Handler": "app.Handler",
    }


@pytest.fixture
def valid_sklearn_properties() -> dict:
    return {
        "Description": "Example description",
        "Name": "autoencoder",
        "MetricKey": "output_1",
        "BuiltinThreshold": False,
        "AnomalyThreshold": 0.1,
        "InputSignature": [
            {
                "Name": "input_0",
                "MonitoringElement": "JetMET/MET/pfMETT1/Cleaned/METSig",
                "DataType": "TYPE_FP32",
                "Dims": [-1, 51],
            }
        ],
        "OutputSignature": [
            {"Name": "output_0", "DataType": "TYPE_FP32", "Dims": [-1, 51]},
            {"Name": "output_1", "DataType": "TYPE_FP32", "Dims": [-1]},
        ],
        "CodeUri": "example/sklearn",
        "ModelUri": "example/sklearn/model.joblib",
        "Handler": "app.Handler",
    }


@pytest.fixture
def valid_custom_image_properties() -> dict:
    return {
        "Description": "Example description",
        "Name": "autoencoder",
        "MetricKey": "output_1",
        "BuiltinThreshold": False,
        "AnomalyThreshold": 0.1,
        "InputSignature": [
            {
                "Name": "input_0",
                "MonitoringElement": "JetMET/MET/pfMETT1/Cleaned/METSig",
                "DataType": "TYPE_FP32",
                "Dims": [-1, 51],
            }
        ],
        "OutputSignature": [
            {"Name": "output_0", "DataType": "TYPE_FP32", "Dims": [-1, 51]},
            {"Name": "output_1", "DataType": "TYPE_FP32", "Dims": [-1]},
        ],
        "CodeUri": "example/sklearn",
        "ModelUri": "example/sklearn/model.joblib",
        "Handler": "app.Handler",
        "Image": "seldonio/mlserver:1.5.0",
    }


@pytest.fixture
def valid_custom_resources_properties() -> dict:
    return {
        "Description": "Example description",
        "Name": "autoencoder",
        "MetricKey": "output_1",
        "BuiltinThreshold": False,
        "AnomalyThreshold": 0.1,
        "InputSignature": [
            {
                "Name": "input_0",
                "MonitoringElement": "JetMET/MET/pfMETT1/Cleaned/METSig",
                "DataType": "TYPE_FP32",
                "Dims": [-1, 51],
            }
        ],
        "OutputSignature": [
            {"Name": "output_0", "DataType": "TYPE_FP32", "Dims": [-1, 51]},
            {"Name": "output_1", "DataType": "TYPE_FP32", "Dims": [-1]},
        ],
        "CodeUri": "example/sklearn",
        "ModelUri": "example/sklearn/model.joblib",
        "Handler": "app.Handler",
        "Resources": {
            "Requests": {"cpu": "500m", "memory": "1Gi", "nvidia.com/gpu": "1"},
            "Limits": {"cpu": "1", "memory": "2Gi", "nvidia.com/gpu": "1"},
        },
    }


def common_assert(properties):
    assert properties.Description == "Example description"
    assert properties.Name == "autoencoder"
    assert properties.MetricKey == "output_1"
    assert properties.BuiltinThreshold is False
    assert properties.AnomalyThreshold == 0.1
    assert len(properties.InputSignature) == 1
    assert len(properties.OutputSignature) == 2
    assert properties.InputSignature[0].Name == "input_0"
    assert properties.InputSignature[0].MonitoringElement == "JetMET/MET/pfMETT1/Cleaned/METSig"
    assert properties.InputSignature[0].DataType == DataTypeEnum.FP32
    assert properties.InputSignature[0].Dims == [-1, 51]
    assert properties.OutputSignature[0].Name == "output_0"
    assert properties.OutputSignature[0].DataType == DataTypeEnum.FP32
    assert properties.OutputSignature[0].Dims == [-1, 51]
    assert properties.OutputSignature[1].Name == "output_1"
    assert properties.OutputSignature[1].DataType == DataTypeEnum.FP32
    assert properties.OutputSignature[1].Dims == [-1]


def test_torchscript(valid_torchcript_properties):
    props = TorchscriptProperties(**valid_torchcript_properties)
    common_assert(props)
    assert props.MaxBatchSize == 0
    assert props.ModelUri == Path("example/model.pt")
    assert props.Platform == PlatformType.TORCHSCRIPT
    with pytest.raises(AttributeError, match="Platform is a constant and cannot be changed."):
        props.Platform = "invalid"


def test_tensorflow_savedmodel(valid_tensorflow_savedmodel_properties):
    props = TensorflowSavedModelProperties(**valid_tensorflow_savedmodel_properties)
    common_assert(props)
    assert props.MaxBatchSize == 0
    assert props.SavedModelUri == Path("example/saved_model")
    assert props.Platform == PlatformType.TENSORFLOW_SAVEDMODEL
    with pytest.raises(AttributeError, match="Platform is a constant and cannot be changed."):
        props.Platform = "invalid"
    props.MaxBatchSize = 10


def test_onnx(valid_onnx_properties):
    props = ONNXProperties(**valid_onnx_properties)
    common_assert(props)
    assert props.MaxBatchSize == 0
    assert props.ModelUri == Path("example/model.onnx")
    assert props.Platform == PlatformType.ONNX
    with pytest.raises(AttributeError, match="Platform is a constant and cannot be changed."):
        props.Platform = "invalid"


def test_xgboost(valid_xgboost_properties):
    props = XGBoostProperties(**valid_xgboost_properties)
    common_assert(props)
    assert props.ModelUri == Path("example/xgboost/0001.ubj")
    assert props.CodeUri == Path("example/xgboost")


def test_lightgbm(valid_lightgbm_properties):
    props = LightGBMProperties(**valid_lightgbm_properties)
    common_assert(props)
    assert props.ModelUri == Path("example/lightgbm/0001.txt")
    assert props.CodeUri == Path("example/lightgbm")


def test_sklearn(valid_sklearn_properties):
    props = SKLearnProperties(**valid_sklearn_properties)
    common_assert(props)
    assert props.ModelUri == Path("example/sklearn/model.joblib")
    assert props.CodeUri == Path("example/sklearn")


def test_custom_image(valid_custom_image_properties):
    props = SKLearnProperties(**valid_custom_image_properties)
    common_assert(props)
    assert props.ModelUri == Path("example/sklearn/model.joblib")
    assert props.CodeUri == Path("example/sklearn")
    assert props.Image == "seldonio/mlserver:1.5.0"


def test_custom_resources(valid_custom_resources_properties):
    props = SKLearnProperties(**valid_custom_resources_properties)
    common_assert(props)
    assert props.ModelUri == Path("example/sklearn/model.joblib")
    assert props.CodeUri == Path("example/sklearn")
    assert props.Resources is not None
    assert props.Resources.Requests.cpu == "500m"
    assert props.Resources.Requests.memory == "1Gi"
    assert props.Resources.Requests.gpu == "1"
    assert props.Resources.Limits.cpu == "1"
    assert props.Resources.Limits.memory == "2Gi"
    assert props.Resources.Limits.gpu == "1"



from typing import Union

from pydantic import BaseModel, Field, model_validator

from ...properties.isvc.mlserver import (
    LightGBMProperties,
    SKLearnProperties,
    XGBoostProperties,
)
from ...properties.isvc.tritonserver import (
    ONNXProperties,
    TensorflowSavedModelProperties,
    TorchscriptProperties,
)
from ..types import ResourceType
from .types import ModelType, ServingFrameworkType


class InferenceServiceResource(BaseModel):
    SuperType: ResourceType
    ServingFrameworkType: ServingFrameworkType
    ModelType: ModelType
    Type: str
    Workspace: str
    Properties: Union[
        TorchscriptProperties,
        TensorflowSavedModelProperties,
        ONNXProperties,
        XGBoostProperties,
        LightGBMProperties,
        SKLearnProperties,
    ] = Field(...)

    @model_validator(mode="before")
    @classmethod
    def validate_and_map_props(cls, resource):
        try:
            super_type, framework, model = resource["Type"].split("::")
        except ValueError as err:
            raise ValueError(
                f"Invalid Type format: {resource['Type']}. Expected format: 'SuperType::Framework::Model'."
            ) from err

        # Validate the components
        if super_type != ResourceType.INFERENCE_SERVICE:
            raise NotImplementedError(f"Unsupported SuperType: {super_type}")
        if framework not in ServingFrameworkType.__members__.values():
            raise NotImplementedError(f"Unsupported ServingFramework: {framework}")
        if model not in ModelType.__members__.values():
            raise NotImplementedError(f"Unsupported ModelType: {model}")

        # Map the resource to the appropriate properties
        if (framework, model) == (ServingFrameworkType.TRITONSERVER, ModelType.TORCHSCRIPT):
            props = TorchscriptProperties(**resource["Properties"])
        elif (framework, model) == (ServingFrameworkType.TRITONSERVER, ModelType.TENSORFLOW_SAVEDMODEL):
            props = TensorflowSavedModelProperties(**resource["Properties"])
        elif (framework, model) == (ServingFrameworkType.TRITONSERVER, ModelType.ONNX):
            props = ONNXProperties(**resource["Properties"])
        elif (framework, model) == (ServingFrameworkType.ML_SERVER, ModelType.XGBOOST):
            props = XGBoostProperties(**resource["Properties"])
        elif (framework, model) == (ServingFrameworkType.ML_SERVER, ModelType.LIGHTGBM):
            props = LightGBMProperties(**resource["Properties"])
        elif (framework, model) == (ServingFrameworkType.ML_SERVER, ModelType.SKLEARN):
            props = SKLearnProperties(**resource["Properties"])
        else:
            raise NotImplementedError(f"Resource of type {resource['Type']} is not supported.")

        # Update the resource with the mapped properties
        resource["SuperType"] = ResourceType(super_type)
        resource["ServingFrameworkType"] = ServingFrameworkType(framework)
        resource["ModelType"] = ModelType(model)
        resource["Properties"] = props
        return resource



import pytest

from dism_core.properties.base import BaseProperties, Tag


@pytest.fixture
def valid_props_no_tags() -> dict:
    return {
        "Description": "Example description",
        "Name": "autoencoder",
    }


@pytest.fixture
def valid_props_with_tags() -> dict:
    return {
        "Description": "Example description",
        "Name": "autoencoder",
        "Tags": [
            {"Name": "tag1", "Value": "value1"},
            {"Name": "tag2", "Value": "value2"},
        ],
    }


def test_valid_props_no_tags(valid_props_no_tags):
    props = BaseProperties(**valid_props_no_tags)
    assert props.Description == "Example description"
    assert props.Name == "autoencoder"
    assert props.Tags is None
    assert isinstance(props, BaseProperties)


def test_valid_props_with_tags(valid_props_with_tags):
    props = BaseProperties(**valid_props_with_tags)
    assert props.Description == "Example description"
    assert props.Name == "autoencoder"
    assert props.Tags is not None
    assert isinstance(props.Tags, list)
    assert all(isinstance(tag, Tag) for tag in props.Tags)
    assert isinstance(props, BaseProperties)


def test_invalid_tag():
    with pytest.raises(
        ValueError, match="String must contain only lowercase alphanumeric characters, hyphens, or underscores"
    ):
        Tag(Name="Invalid Tag", Value="value1")


def test_invalid_name():
    with pytest.raises(ValueError, match="String violates RFC 1123 *"):
        BaseProperties(Description="Example description", Name="Invalid Name")
    with pytest.raises(ValueError, match="String violates RFC 1123 *"):
        BaseProperties(Description="Example description", Name="Autoencoder")



import re
from typing import Annotated, Optional

from pydantic import BaseModel, StringConstraints, field_validator


class Tag(BaseModel):
    Name: Annotated[
        str, StringConstraints(max_length=63)
    ]  # There are not hard restrictions for this, but let's keep things standardized
    Value: str

    @field_validator("Name")
    @classmethod
    def validate_name_chars(cls, v):
        if not re.fullmatch(r"^[a-z0-9\-_]+$", v):
            raise ValueError("String must contain only lowercase alphanumeric characters, hyphens, or underscores")
        return v


class BaseProperties(BaseModel):
    Description: str
    Name: Annotated[
        str, StringConstraints(max_length=36)
    ]  # The final service name has to be limited to 63 characters, it will be a composed string: {stage:5}-{workspace:18}-{Name:36}-{version:4}
    Tags: Optional[list[Tag]] = None

    @field_validator("Name")
    @classmethod
    def validate_name_chars(cls, v):
        pattern = r"^[a-z0-9]([-a-z0-9]*[a-z0-9])?(\.[a-z0-9]([-a-z0-9]*[a-z0-9])?)*$"
        if not re.fullmatch(pattern, v):
            raise ValueError(f"String violates RFC 1123 subdomain naming conventions. Validation is done by {pattern}")
        return v



from pathlib import Path

from .base import TritonserverProperties
from .types import PlatformType


class TensorflowSavedModelProperties(TritonserverProperties):
    SavedModelUri: Path
    Platform: PlatformType = PlatformType.TENSORFLOW_SAVEDMODEL



from typing import Optional

from pydantic import BaseModel, ConfigDict, Field


class ResourceSpec(BaseModel):
    cpu: Optional[str] = None
    memory: Optional[str] = None
    gpu: Optional[str] = Field(default=None, alias="nvidia.com/gpu")

    model_config = ConfigDict(populate_by_name=True)


class Resources(BaseModel):
    Requests: Optional[ResourceSpec] = None
    Limits: Optional[ResourceSpec] = None



import logging

from .template import Template
from .validation.isvc import InferenceServiceValidator


logger = logging.getLogger(__name__)


try:
    from ._version import __version__, __version_tuple__
except ImportError:
    logger.warning(
        "The correct version of dism-core could not be resolved. This is likely due to "
        "a broken installation. The package needs to be correctly built, so that "
        "the build-system properly generates the version file. "
    )
    __version__ = "unknown"
    __version_tuple__ = (0, 0, "unknown")


__all__ = [
    "InferenceServiceValidator",
    "Template",
    "__version__",
    "__version_tuple__",
]



from enum import Enum


class ServingFrameworkType(str, Enum):
    TRITONSERVER = "Tritonserver"
    ML_SERVER = "MLServer"


class ModelType(str, Enum):
    TORCHSCRIPT = "Torchscript"
    TENSORFLOW_SAVEDMODEL = "TensorflowSavedModel"
    ONNX = "ONNX"
    XGBOOST = "XGBoost"
    LIGHTGBM = "LightGBM"
    SKLEARN = "SKLearn"



from pathlib import Path

from .base import TritonserverProperties
from .types import PlatformType


class TorchscriptProperties(TritonserverProperties):
    ModelUri: Path
    Platform: PlatformType = PlatformType.TORCHSCRIPT



import json
from pathlib import Path
from typing import Any, Union

import yaml
from pydantic import BaseModel, Field, model_validator

from .resources.isvc import InferenceServiceResource
from .resources.types import ResourceType
from .versions import TemplateFormatVersion


class Template(BaseModel):
    TemplateFormatVersion: TemplateFormatVersion
    Description: str
    Resources: dict[str, InferenceServiceResource] = Field(...)

    @staticmethod
    def validate_version_1(values: dict[str, Any]) -> dict[str, Any]:
        resources = values.get("Resources", {})
        mapped_resources = {}
        for key, resource in resources.items():
            try:
                super_type, _, _ = resource["Type"].split("::")
            except ValueError as err:
                raise ValueError(
                    f"Invalid Type format: {resource['Type']}. Expected format: 'SuperType::Framework::Model'."
                ) from err

            # Validate the components
            if super_type != ResourceType.INFERENCE_SERVICE:
                raise ValueError(f"Unsupported SuperType: {super_type}")

            if super_type == ResourceType.INFERENCE_SERVICE.value:
                mapped_resources[key] = InferenceServiceResource(**resource)
            else:
                raise NotImplementedError(f"Resource of type {resource['Type']} is not supported.")

        values["Resources"] = mapped_resources
        return values

    @model_validator(mode="before")
    @classmethod
    def validate_and_map_resource(cls, values: dict[str, dict]) -> dict[str, InferenceServiceResource]:
        version = TemplateFormatVersion(values.get("TemplateFormatVersion"))
        if version == TemplateFormatVersion.VERSION_1:
            return Template.validate_version_1(values)

    @classmethod
    def from_json_file(cls, file_path: Union[str, Path]) -> "Template":
        with open(file_path) as file:
            data = json.load(file)
        return cls(**data)

    @classmethod
    def from_json_str(cls, json_string: str) -> "Template":
        data = json.loads(json_string)
        return cls(**data)

    @classmethod
    def from_yaml_file(cls, file_path: Union[str, Path]) -> "Template":
        with open(file_path) as file:
            data = yaml.safe_load(file)
        return cls(**data)

    @classmethod
    def from_yaml_str(cls, yaml_string: str) -> "Template":
        data = yaml.safe_load(yaml_string)
        return cls(**data)

    def to_dict(self) -> dict[str, Any]:
        return self.model_dump(mode="json")

    def to_json(self, fpath: Union[str, Path]) -> None:
        with open(fpath, "w") as file:
            json.dump(self.to_dict(), file, indent=4)

    def to_yaml(self, fpath: Union[str, Path]) -> None:
        with open(fpath, "w") as file:
            yaml.safe_dump(self.to_dict(), file)



from pathlib import Path
from unittest.mock import mock_open, patch

import pytest

from dism_core import InferenceServiceValidator, Template
from dism_core.validation.utils import has_handler_class


@pytest.fixture
def valid_yaml():
    return """
TemplateFormatVersion: "2025-03-31"
Description: Deploys multiple models to spot anomalies in the JME METSig distribution
Resources:
  MetSigPyTorchAutoEncoder:
    Type: InferenceService::Tritonserver::Torchscript
    Workspace: jetmet
    Properties:
      Description: Spot anomalies in the MetSig distribution
      Name: metsig-torchscript-autoencoder
      MetricKey: output_1
      BuiltinThreshold: false
      AnomalyThreshold: 0.1
      InputSignature:
        - Name: input_0
          MonitoringElement: JetMET/MET/pfMETT1/Cleaned/METSig
          DataType: TYPE_FP32
          Dims:
            - -1
            - 51
      OutputSignature:
        - Name: output_0
          DataType: TYPE_FP32
          Dims:
            - -1
            - 51
        - Name: output_1
          DataType: TYPE_FP32
          Dims:
            - -1
      MaxBatchSize: 0
      ModelUri: examples/mldeploy/torchscript/model.pt
  MetSigTensorflowAutoEncoder:
    Type: InferenceService::Tritonserver::TensorflowSavedModel
    Workspace: jetmet
    Properties:
      Description: Spot anomalies in the MetSig distribution
      Name: metsig-tensorflow-autoencoder
      MetricKey: output_1
      BuiltinThreshold: false
      AnomalyThreshold: 0.1
      InputSignature:
        - Name: input_0
          MonitoringElement: JetMET/MET/pfMETT1/Cleaned/METSig
          DataType: TYPE_FP32
          Dims:
            - -1
            - 51
      OutputSignature:
        - Name: output_0
          DataType: TYPE_FP32
          Dims:
            - -1
            - 51
        - Name: output_1
          DataType: TYPE_FP32
          Dims:
            - -1
      MaxBatchSize: 0
      SavedModelUri: examples/mldeploy/tensorflow/
  MetSigOnnxAutoEncoder:
    Type: InferenceService::Tritonserver::ONNX
    Workspace: jetmet
    Properties:
      Description: Spot anomalies in the MetSig distribution
      Name: metsig-onnx-autoencoder
      MetricKey: output_1
      BuiltinThreshold: false
      AnomalyThreshold: 0.1
      InputSignature:
        - Name: input_0
          MonitoringElement: JetMET/MET/pfMETT1/Cleaned/METSig
          DataType: TYPE_FP32
          Dims:
            - -1
            - 51
      OutputSignature:
        - Name: output_0
          DataType: TYPE_FP32
          Dims:
            - -1
            - 51
        - Name: output_1
          DataType: TYPE_FP32
          Dims:
            - -1
      MaxBatchSize: 0
      ModelUri: examples/mldeploy/onnx/model.onnx
  MetSigXGBoostRegressor:
    Type: InferenceService::MLServer::XGBoost
    Workspace: jetmet
    Properties:
      Description: Spot anomalies in the MetSig distribution
      Name: metsig-xgboost-autoencoder
      MetricKey: avg_mse
      BuiltinThreshold: false
      AnomalyThreshold: 0.1
      InputSignature:
        - Name: input_0
          MonitoringElement: JetMET/MET/pfMETT1/Cleaned/METSig
          DataType: TYPE_FP32
          Dims:
            - -1
            - 51
      OutputSignature:
        - Name: reconstruction
          DataType: TYPE_FP32
          Dims:
            - -1
            - 51
        - Name: avg_mse
          DataType: TYPE_FP32
          Dims:
            - -1
      CodeUri: examples/mldeploy/xgboost/
      Handler: app.Handler
      ModelUri: examples/mldeploy/xgboost/0001.ubj
  MetSigLightGBMRegressor:
    Type: InferenceService::MLServer::SKLearn
    Workspace: jetmet
    Properties:
      Description: Spot anomalies in the MetSig distribution
      Name: metsig-lightgbm-autoencoder
      MetricKey: avg_mse
      BuiltinThreshold: false
      AnomalyThreshold: 0.1
      InputSignature:
        - Name: input_0
          MonitoringElement: JetMET/MET/pfMETT1/Cleaned/METSig
          DataType: TYPE_FP32
          Dims:
            - -1
            - 51
      OutputSignature:
        - Name: reconstruction
          DataType: TYPE_FP32
          Dims:
            - -1
            - 51
        - Name: avg_mse
          DataType: TYPE_FP32
          Dims:
            - -1
      CodeUri: examples/mldeploy/sklearn-lightgbm/
      Handler: app.Handler
      ModelUri: examples/mldeploy/sklearn-lightgbm/model.joblib
  MetSigNMFRegressor:
    Type: InferenceService::MLServer::SKLearn
    Workspace: jetmet
    Properties:
      Description: Spot anomalies in the MetSig distribution
      Name: metsig-nmf-autoencoder
      MetricKey: avg_mse
      BuiltinThreshold: false
      AnomalyThreshold: 0.1
      InputSignature:
        - Name: input_0
          MonitoringElement: JetMET/MET/pfMETT1/Cleaned/METSig
          DataType: TYPE_FP32
          Dims:
            - -1
            - 51
      OutputSignature:
        - Name: reconstruction
          DataType: TYPE_FP32
          Dims:
            - -1
            - 51
        - Name: avg_mse
          DataType: TYPE_FP32
          Dims:
            - -1
      CodeUri: examples/mldeploy/sklearn-nmf/
      Handler: app.Handler
      ModelUri: examples/mldeploy/sklearn-nmf/model.joblib
    """


@pytest.fixture
def template(valid_yaml):
    return Template.from_yaml_str(valid_yaml)


@pytest.fixture
def app_script():
    return """
import numpy as np
import xgboost as xgb
from datatype import datatype_to_dtype, dtype_to_datatype
from mlserver import MLModel
from mlserver.types import InferenceRequest, InferenceResponse, RequestInput, RequestOutput, ResponseOutput
from mlserver.utils import get_model_uri
from sklearn.preprocessing import MinMaxScaler


class Handler(MLModel):
    async def load(self):
        model_uri = await get_model_uri(self._settings)
        self.model_name = self._settings.name
        self.model_version = self._settings.version
        self.model = xgb.Booster(model_file=model_uri)
"""


def test_valid_uris(template, app_script):
    validator = InferenceServiceValidator(template)
    with patch("builtins.open", mock_open(read_data=app_script)):
        with patch.object(Path, "is_file", return_value=True):
            with patch.object(Path, "is_dir", return_value=True):
                validator()


def test_file_not_found(template, app_script):
    validator = InferenceServiceValidator(template)
    with patch("builtins.open", mock_open(read_data=app_script)):
        with patch.object(Path, "is_file", return_value=False):
            with patch.object(Path, "is_dir", return_value=True):
                with pytest.raises(ValueError, match="File .* not found or is not a file."):
                    validator()


def test_dir_not_found(template, app_script):
    validator = InferenceServiceValidator(template)
    with patch("builtins.open", mock_open(read_data=app_script)):
        with patch.object(Path, "is_file", return_value=True):
            with patch.object(Path, "is_dir", return_value=False):
                with pytest.raises(ValueError, match="Directory .* not found or is not a directory."):
                    validator()


def test_has_handler_class_not_found():
    mock_file_content = """
import something

class NotMyHandler(BaseHandler):
    pass
    """
    with patch("builtins.open", mock_open(read_data=mock_file_content)):
        result = has_handler_class("fake/path/to/handler.py", "MyHandler")
        assert result is False


def test_has_handler_class_with_spacing_variations():
    mock_file_content = """
class    MyHandler   (BaseHandler):
    pass
    """
    with patch("builtins.open", mock_open(read_data=mock_file_content)):
        result = has_handler_class("fake/path.py", "MyHandler")
        assert result is True



import pytest

from dism_core.properties.isvc.tritonserver.torchscript import TorchscriptProperties
from dism_core.resources.isvc import InferenceServiceResource
from dism_core.resources.isvc.types import ModelType, ServingFrameworkType
from dism_core.resources.types import ResourceType


@pytest.fixture
def valid_resource() -> dict:
    return {
        "Type": "InferenceService::Tritonserver::Torchscript",
        "Workspace": "jetmet",
        "Properties": {
            "Description": "Example description",
            "Name": "autoencoder",
            "MetricKey": "output_1",
            "BuiltinThreshold": False,
            "AnomalyThreshold": 0.1,
            "InputSignature": [
                {
                    "Name": "input_0",
                    "MonitoringElement": "JetMET/MET/pfMETT1/Cleaned/METSig",
                    "DataType": "TYPE_FP32",
                    "Dims": [-1, 51],
                }
            ],
            "OutputSignature": [
                {"Name": "output_0", "DataType": "TYPE_FP32", "Dims": [-1, 51]},
                {"Name": "output_1", "DataType": "TYPE_FP32", "Dims": [-1]},
            ],
            "MaxBatchSize": 0,
            "ModelUri": "example/model.pt",
        },
    }


def test_resource(valid_resource):
    resource = InferenceServiceResource(**valid_resource)
    assert resource.SuperType == ResourceType.INFERENCE_SERVICE
    assert resource.ServingFrameworkType == ServingFrameworkType.TRITONSERVER
    assert resource.ModelType == ModelType.TORCHSCRIPT
    assert resource.Type == "InferenceService::Tritonserver::Torchscript"
    assert resource.Workspace == "jetmet"
    assert isinstance(resource.Properties, TorchscriptProperties)


def test_invalid_resource():
    with pytest.raises(
        NotImplementedError, match="Resource of type InferenceService::MLServer::Torchscript is not supported."
    ):
        InferenceServiceResource(
            **{"Type": "InferenceService::MLServer::Torchscript", "Workspace": "jetmet", "Properties": {}}
        )



import json
from unittest.mock import mock_open, patch

import pytest
import yaml
from pydantic import ValidationError

from dism_core import Template
from dism_core.resources.isvc.types import ModelType, ServingFrameworkType
from dism_core.resources.types import ResourceType
from dism_core.versions import TemplateFormatVersion


@pytest.fixture
def valid_yaml():
    return """
TemplateFormatVersion: "2025-03-31"
Description: Example template
Resources:
    MyResource:
        Type: InferenceService::Tritonserver::Torchscript
        Workspace: test
        Properties:
            Description: Example description
            Name: autoencoder
            MetricKey: output_1
            BuiltinThreshold: false
            AnomalyThreshold: 0.1
            InputSignature:
            -
                Name: input_0
                MonitoringElement: JetMET/MET/pfMETT1/Cleaned/METSig
                DataType: TYPE_FP32
                Dims:
                    - -1
                    - 51
            OutputSignature:
            -
                Name: output_0
                DataType: TYPE_FP32
                Dims:
                    - -1
                    - 51
            -
                Name: output_1
                DataType: TYPE_FP32
                Dims:
                    - -1
            MaxBatchSize: 0
            ModelUri: example/model.pt
    """


@pytest.fixture
def valid_json():
    return """
    {
        "TemplateFormatVersion": "2025-03-31",
        "Description": "Example template",
        "Resources": {
            "MyResource": {
                "Type": "InferenceService::Tritonserver::Torchscript",
                "Workspace": "jetmet",
                "Properties": {
                    "Description": "Example description",
                    "Name": "autoencoder",
                    "MetricKey": "output_1",
                    "BuiltinThreshold": false,
                    "AnomalyThreshold": 0.1,
                    "InputSignature": [
                        {
                            "Name": "input_0",
                            "MonitoringElement": "JetMET/MET/pfMETT1/Cleaned/METSig",
                            "DataType": "TYPE_FP32",
                            "Dims": [-1, 51]
                        }
                    ],
                    "OutputSignature": [
                        {
                            "Name": "output_0",
                            "DataType": "TYPE_FP32",
                            "Dims": [-1, 51]
                        },
                        {
                            "Name": "output_1",
                            "DataType": "TYPE_FP32",
                            "Dims": [-1]
                        }
                    ],
                    "MaxBatchSize": 0,
                    "ModelUri": "example/model.pt"
                }
            }
        }
    }
    """


def common_assert(template):
    assert template.TemplateFormatVersion == TemplateFormatVersion.VERSION_1
    assert template.Description == "Example template"
    assert "MyResource" in template.Resources
    assert template.Resources["MyResource"].SuperType == ResourceType.INFERENCE_SERVICE
    assert template.Resources["MyResource"].ServingFrameworkType == ServingFrameworkType.TRITONSERVER
    assert template.Resources["MyResource"].ModelType == ModelType.TORCHSCRIPT
    assert template.Resources["MyResource"].Type == "InferenceService::Tritonserver::Torchscript"


def test_from_yaml_str_valid(valid_yaml):
    template = Template.from_yaml_str(valid_yaml)
    common_assert(template)


def test_from_json_str_valid(valid_json):
    template = Template.from_json_str(valid_json)
    common_assert(template)


def test_invalid_yaml_format():
    invalid_yaml = """
    TemplateFormatVersion: "1.0"
    Description: "Example template"
    Resources:
      MyResource:
        Type: "InvalidType"
    """
    with pytest.raises(ValidationError, match="validation error for Template"):
        Template.from_yaml_str(invalid_yaml)


def test_invalid_json_format():
    invalid_json = """
    {
        "TemplateFormatVersion": "1.0",
        "Description": "Example template",
        "Resources": {
            "MyResource": {
                "Type": "InvalidType"
            }
        }
    }
    """
    with pytest.raises(ValidationError, match="validation error for Template"):
        Template.from_json_str(invalid_json)


@patch("builtins.open", new_callable=mock_open)
def test_from_yaml_file_valid(mock_file, valid_yaml):
    mock_file.return_value.__enter__.return_value = valid_yaml
    template = Template.from_yaml_file("template.yaml")
    common_assert(template)


@patch("builtins.open", new_callable=mock_open)
def test_from_json_file_valid(mock_file, valid_json):
    mock_file.return_value.read.return_value = valid_json
    template = Template.from_json_file("template.json")
    common_assert(template)


@patch("builtins.open", new_callable=mock_open)
def test_file_not_found(mock_file):
    mock_file.side_effect = FileNotFoundError
    with pytest.raises(FileNotFoundError):
        Template.from_yaml_file("nonexistent.yaml")
    with pytest.raises(FileNotFoundError):
        Template.from_json_file("nonexistent.json")


def test_dictify(valid_yaml):
    template = Template.from_yaml_str(valid_yaml)
    template_dict = template.to_dict()
    assert isinstance(template_dict, dict)
    assert template_dict["TemplateFormatVersion"] == TemplateFormatVersion.VERSION_1.value
    assert template_dict["Description"] == "Example template"
    assert isinstance(template_dict["Resources"], dict)
    assert "MyResource" in template_dict["Resources"]
    assert template_dict["Resources"]["MyResource"]["SuperType"] == ResourceType.INFERENCE_SERVICE.value
    assert template_dict["Resources"]["MyResource"]["ServingFrameworkType"] == ServingFrameworkType.TRITONSERVER.value
    assert template_dict["Resources"]["MyResource"]["ModelType"] == ModelType.TORCHSCRIPT.value
    assert template_dict["Resources"]["MyResource"]["Type"] == "InferenceService::Tritonserver::Torchscript"


@patch("builtins.open", new_callable=mock_open)
def test_to_json(mock_file, valid_yaml):
    template = Template.from_yaml_str(valid_yaml)
    template.to_json("output.json")
    mock_file.assert_called_once_with("output.json", "w")
    call_args_list = mock_file().write.call_args_list
    written_data = "".join(call.args[0] for call in call_args_list)
    assert json.loads(written_data) == template.to_dict()


@patch("builtins.open", new_callable=mock_open)
def test_to_yaml(mock_file, valid_yaml):
    template = Template.from_yaml_str(valid_yaml)
    template.to_yaml("output.yaml")
    mock_file.assert_called_once_with("output.yaml", "w")
    call_args_list = mock_file().write.call_args_list
    written_data = "".join(call.args[0] for call in call_args_list)
    assert yaml.safe_load(written_data) == template.to_dict()



from typing import Optional

from ..base import BaseProperties
from .resources import Resources as ResourcesModel
from .signature import InputSignature, OutputSignature


class InferenceServiceProperties(BaseProperties):
    AnomalyThreshold: Optional[float] = None
    BuiltinThreshold: Optional[bool] = None
    FlaggingKey: Optional[str] = None
    InputSignature: list[InputSignature]
    MetricKey: str
    OutputSignature: list[OutputSignature]
    Image: Optional[str] = None
    Resources: Optional[ResourcesModel] = None



from .base import TritonserverProperties
from .onnx import ONNXProperties
from .tensorflow_savedmodel import TensorflowSavedModelProperties
from .torchscript import TorchscriptProperties


__all__ = [
    "ONNXProperties",
    "TensorflowSavedModelProperties",
    "TorchscriptProperties",
    "TritonserverProperties",
]



from .base import MLServerProperties
from .lightgbm import LightGBMProperties
from .sklearn import SKLearnProperties
from .xgboost import XGBoostProperties


__all__ = [
    "LightGBMProperties",
    "MLServerProperties",
    "SKLearnProperties",
    "XGBoostProperties",
]



import re
from pathlib import Path
from typing import Union


def has_handler_class(file_path: Union[str, Path], class_name: str) -> bool:
    pattern = re.compile(rf"(?m)^\s*class\s+{class_name}\s*\(")
    with open(file_path) as file:
        for line in file:
            if pattern.search(line):
                return True
    return False
